{-# LANGUAGE NoImplicitPrelude #-}
{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE DerivingVia #-}
{-# LANGUAGE TupleSections #-}
{-# OPTIONS_GHC -Wno-unrecognised-pragmas #-}
module Chart (newChart
             , FillerException(..)
             , getUpdatedChartCache
             , getChunkedDay
             , timePointToIndex
             , indexToTimePoint
             ) where

import Import
import Control.Concurrent (forkIO)
import Control.Monad.Extra (ifM)
import Data.Aeson (eitherDecodeFileStrict')
import Data.FingerTree (FingerTree)
import Data.Time.Calendar (Day(..))
import Data.Time.Clock
import Data.Maybe (fromJust)
import RIO.FilePath
import RIO.List
import RIO.List.Partial

import qualified Data.FingerTree as FT
import qualified Data.HashMap.Strict as HM
import qualified Data.List as L
import qualified Data.Text as T
import qualified Data.Vector.Storable as VS

import qualified Debug.Trace as D

data FillerException = QuitFiller
  deriving stock Show

instance Exception FillerException

newChart :: IBContract -> RIO App ()
newChart contract = do
  app <- ask
  let sym = (symbol :: IBContract -> Text) contract
      con = (conId :: IBContract -> Int) contract
      hmVar = appCharts . appRefs $ app
  hm <- liftIO . readTVarIO $ hmVar
  unless (sym `HM.member` hm) $ do
    today <- liftIO getCurrentDay :: RIO App Day
    (cacheErrors, cacheData) <- do
        ifM (fmap not $ liftIO $ doesDirectoryExist $ "cache" </> show con)
            (return $ (["no data-chache found for " <> show con], [])) $ do
                files <- liftIO $ listDirectory $ "cache" </> show con
                res <- forM files $ \cacheFileName -> do
                  let fname = "cache" </> show con </> cacheFileName
                  ifM (liftIO $ doesFileExist fname)
                      (bimap (\err -> cacheFileName <> ": "<> err) ((readMaybe $ dropExtension cacheFileName) :: Maybe Day,) <$> liftIO (eitherDecodeFileStrict' fname))
                      (return $ Left $ "cachefile "<>cacheFileName<>" not found.")
                return $ partitionEithers res
    unless (null cacheErrors) $ forM_ cacheErrors $ \err -> logError (displayShow err)
    let (unknownDates, cacheData') = partition (isNothing . fst) cacheData
    unless (null unknownDates) $ logError $ display $ T.pack $ show (length unknownDates) <> " files had no parsable Date."
    let cacheData'' :: HashMap Day (FingerTree TimeInterval ChartPoint) = HM.fromList $ bimap (fromMaybe today) (FT.fromList . L.sortOn (intervalFrom . timeOfDay)) <$> cacheData'
    logError $ displayShow $ HM.keys cacheData''
    logInfo $ displayShow(fromMaybe FT.empty $ cacheData'' HM.!? today)
    c <- liftIO $ newTVarIO $ Chart con
                                    (fromMaybe FT.empty $ cacheData'' HM.!? today)
                                    (HM.delete today cacheData'')
                                    undefined
                                    defChartSettings
                                    defChartCacheSettings
                                    (emptyChartCacheData defChartCacheSettings)
                                    Nothing
                                    False
    tid <- liftIO $ forkIO $ fillChart app contract c
    liftIO $ atomically $ do
      modifyTVar' c (\Chart{..} -> let fillerThread = tid in Chart{..})
      modifyTVar' hmVar (HM.insert sym c)

fillChart :: App -> IBContract -> TVar Chart -> IO ()
fillChart app contract cVar = runRIO app $ do
  let sym = (symbol :: IBContract -> Text) contract
  (tickerMapVar :: TVar (HashMap Int Text)) <- tickerIdToSymbol. appRefs <$> ask
  alreadyAdded <- (sym `elem`) . HM.elems <$> liftIO (readTVarIO tickerMapVar)
  unless alreadyAdded $ do
    tickerId <- (+1) . foldl' max 1000 . HM.keys <$> liftIO (readTVarIO tickerMapVar)
    let cancelSubscription = liftIO $ atomically $ do
         modifyTVar tickerMapVar (HM.delete tickerId)
         -- TODO: send cancel-request
    let sendQ = twsConnectionSend . twsConnectionRefs . appRefs $ app
    liftIO $ atomically $ do
      modifyTVar tickerMapVar (HM.insert tickerId sym)
      writeTQueue sendQ $ Msg_IB_OUT $ IB_RequestMktData tickerId contract "233" False False
    handle (\QuitFiller -> cancelSubscription >> exitSuccess) $
      forever $ do
        -- chart dirty? set clean & begin work
        c <- liftIO (readTVarIO cVar)
        when (chartDirty c) $ do
          liftIO $ atomically $ modifyTVar cVar (\c' -> c' { chartDirty = False })
          (lUpdate, cachePoints) <- getUpdatedChartCache c Nothing
          liftIO $ atomically $ modifyTVar cVar (\c' -> c' { chartCache = cachePoints, lastCacheUpdate = lUpdate })
        threadDelay 1000000 -- sleep 5 seconds

chunkChart :: Int -> Int -> Int -> FingerTree TimeInterval ChartPoint -> [(TimeInterval,[ChartPoint])]
chunkChart from until ticks tree = go from ticks interval
    where
      traceShowCommentId' s a   = a -- traceShowCommentId s a
      traceShowComment'   s a b = b -- traceShowComment s a b
      interval = FT.takeUntil (\(TimeInterval x y) -> x > until)
               . FT.dropUntil (\(TimeInterval x y) -> x >= from || (x <= from && y <= until && y >= from))
               $ tree
      go f i t
        | f >= until = traceShowCommentId' "nosplit OverNow" []
        | otherwise       = let searchpred start@(TimeInterval x y) end@(TimeInterval a b)
                                              | start == mempty = False
                                              | end == mempty = True
                                              | otherwise = traceShowComment' "search" (f, x, start, end) $ f+i <= x || f+2*i <= y
                            in case FT.search searchpred t
                            of pos@(FT.Position l x@(ChartPoint{..}) r) 
                                    | l == mempty && r == mempty && (timeIntervalBegin timeOfDay) < f+i-1 -> (traceShowCommentId' "only 1 left" $ (TimeInterval f (f+i-1), [x])) : go (f+i) i mempty
                                    | l == mempty -> (traceShowComment' ("empty (" <> show f <> "," <> show (f+i) <> ")") x (TimeInterval f (f+i-1), [])) : go (f+i) i (x FT.<| r)
                                    | otherwise -> (traceShowComment' ("split (" <> show f <> "," <> show (f+i) <> ")") pos (FT.measure l,toList l)) : traceShowComment' "recursive call with " (f+i, i, x FT.<| r) (go (f+i) i (x FT.<| r))
                               x -> if FT.measure t == mempty then (TimeInterval f (f+i-1), []) : go (f+i) i t else traceShowCommentId' ("nosplit EndOfTree - "<> show x) [(FT.measure t, toList t)]

-- | converts stuff returned from chunkChart into ([1 point per chunks], last point seen in the input)
toChartData :: [ChartStudyType] -> [(TimeInterval, [ChartPoint])] -> (Maybe ChartPoint, [ChartPoint])
toChartData studytypes chunkedChart = (lastDataPoint, foldedData)
    where (!foldedData,lastDataPoint,_) = foldl' folder ([], Nothing, mempty) chunkedChart
          folder :: ([ChartPoint], Maybe ChartPoint, HM.HashMap Int [Float]) -> (TimeInterval, [ChartPoint]) -> ([ChartPoint], Maybe ChartPoint, HM.HashMap Int [Float])
          folder (acc, Nothing       , smaMap) (tp                   , [])   = (ChartPoint tp 0 0 studies:acc, Nothing, smaMap)
            where
              studies = catMaybes $ studytypes <&> \case
                  ChartStudyTypeOpen   -> Just $ OLHC 0 0 0 0
                  ChartStudyTypeHigh   -> Nothing
                  ChartStudyTypeLow    -> Nothing
                  ChartStudyTypeClose  -> Nothing
                  ChartStudyTypeDirect -> Nothing
                  ChartStudyTypeSMA w  -> Nothing --TODO: fixme #13, do the calculation & use data inside smaMap and return (SMA x y, smaUpdate)
                  ChartStudyTypeVolume -> Just $ Volume 0
          folder (acc, Just lastPoint, smaMap) (tp                   , [])   = (lastPoint{ timeOfDay = tp, pointVolume = 0}:acc, Just lastPoint, foldl' (.) id smaUpdates smaMap)
            where
              (studies, smaUpdates) = unzip $ catMaybes $ studytypes <&> \case
                  ChartStudyTypeOpen   -> fmap (\OLHC{..} -> (OLHC olhc_close olhc_close olhc_close olhc_close, id)) . L.find (\case OLHC{..} -> True; _ -> False) . pointExtra $ lastPoint
                  ChartStudyTypeHigh   -> Nothing
                  ChartStudyTypeLow    -> Nothing
                  ChartStudyTypeClose  -> Nothing
                  ChartStudyTypeDirect -> Nothing
                  ChartStudyTypeSMA w  -> Nothing --TODO: fixme #13, do the calculation & use data inside smaMap and return (SMA x y, smaUpdate)
                  ChartStudyTypeVolume -> Just (Volume 0,id)
          folder (acc, _             , smaMap) (tp@(TimeInterval f t),cdata) = (cp:acc, Just cp, foldl' (.) id smaUpdates smaMap)
            where
              cp = ChartPoint tp m vol studies
              (studies, smaUpdates) = unzip $ catMaybes $ studytypes <&> \case
                  ChartStudyTypeOpen   -> Just (OLHC o l h c, id)
                  ChartStudyTypeHigh   -> Nothing
                  ChartStudyTypeLow    -> Nothing
                  ChartStudyTypeClose  -> Nothing
                  ChartStudyTypeDirect -> Nothing
                  ChartStudyTypeSMA w  -> Nothing --TODO: fixme #13, do the calculation & use data inside smaMap and return (SMA x y, smaUpdate)
                  ChartStudyTypeVolume -> Just (Volume vol, id)
              vol = sum $ pointVolume <$> cdata
              as' = pointValue <$> cdata
              ms' = sum $ (\x -> pointValue x * pointVolume x) <$> cdata
              m = ms' / vol
              o = head as'
              c = last as'
              l = minimum as'
              h = maximum as'

getUpdatedChartCache :: Chart -> Maybe ChartCacheSettings -> RIO App (Maybe TimePoint, ChartCacheData)
getUpdatedChartCache Chart{..} chartCacheSettings' = do
    let ccs@(ChartCacheSettings cRes cTicks) = fromMaybe chartCacheSettings chartCacheSettings'
    now <- liftIO getCurrentTime
    let cacheUpdateEnd   = ((utcTimeToSeconds now + cRes) `div` cRes) * cRes
        ChartCacheData ccData ccAxis ccRange ccFill = chartCache
    let traceShowCommentId' s a   = a -- traceShowCommentId s a
        traceShowComment'   s a b = b -- traceShowComment s a b
        lUpdate = utcTimeToSeconds now
        -- - calculate Range and need for shift
        shiftNeccessary = case ccRange of
                            Nothing     -> False
                            Just (TimeInterval _ ma) -> cacheUpdateEnd > ma
        (newRange, shiftInterval) = if not shiftNeccessary && isJust ccRange then (fromJust ccRange, 0) else 
                      let interval = cTicks `div` 20 :: Int
                          iTo = ((cacheUpdateEnd + cRes * interval) `div` (cRes * interval)) * (cRes * interval)
                          iFrom = iTo - 20 * (cRes * interval)
                          shift = case ccRange of
                                  Nothing -> 0
                                  Just (TimeInterval _ ma) -> (iTo - ma) `div` cRes
                      in (TimeInterval iFrom iTo, shift)
        ccAxis' = if not shiftNeccessary && isJust ccRange then ccAxis
                  else VS.enumFromStepN (fromIntegral $ timeIntervalBegin newRange) (fromIntegral cRes) cTicks
    let timePointToIndex' :: TimePoint -> Either String Int
        timePointToIndex' = timePointToIndex (Just $ ChartCacheSettings cRes cTicks) newRange
    let indexToTimePoint' :: Int -> TimePoint
        indexToTimePoint' = indexToTimePoint (Just $ ChartCacheSettings cRes cTicks) newRange
    when (Just newRange /= ccRange) $ logDebug $ displayShow ("range changed:", newRange, ccRange)
    logDebug $ displayShow ("shift?" :: Text, shiftNeccessary, shiftInterval)
    logDebug $ displayShow ("now, updateEnd, lUpdate" :: Text, utcTimeToSeconds now, cacheUpdateEnd, lUpdate)
    let cacheUpdateStart = max (timeIntervalBegin newRange) (cacheUpdateEnd - (cRes * (cTicks+1))) -- 1 more to get history running inside the chunk-fold for fallbacks in cache if we get no data.
    -- create data for updates
    let chunks = case (,) <$> ccFill <*> ccRange of
                  Nothing -> [(cacheUpdateStart, cacheUpdateEnd)]
                  Just ((mi, ma), oldRange) -> let itp = (\(TimePoint p) -> p) . indexToTimePoint (Just ccs) oldRange in if
                        -- Interval mi-ma already updated. get remaining intervals
                        -- mi  ma  cUS cUE
                      | itp ma < cacheUpdateStart -> [(cacheUpdateStart, cacheUpdateEnd)]
                        -- mi?  cUS  mi?  ma  cUE -> [ma,cUE] + rest
                      | itp ma < cacheUpdateEnd -> (<> [(itp ma,cacheUpdateEnd)]) if
                          -- cUS  mi  ma
                          | itp mi > cacheUpdateStart -> [(cacheUpdateStart, itp mi)]
                          | otherwise -> []
                      | otherwise -> []
        -- - chunk them with chunhChart
        chunkedChart = L.concat $ chunks <&> \(start, end) -> chunkChart (max 0 start) end cRes chartData
    let (lastDataPoint, !foldedData') = toChartData (chartStudySettings chartSettings) (traceShowCommentId' "chunkedChart" chunkedChart)
        tpToTi (TimeInterval f t) = case timePointToIndex' (TimePoint (t+f) `div` 2) of
                                       Left e   -> error $ "BUG in Chart.hs. Impossible: " <> e
                                       Right ti -> ti
        -- aggregate data with index into the vectors
        foldedData = (\cp@ChartPoint{..} -> (tpToTi timeOfDay, cp)) <$> (traceShowCommentId' "foldedData" foldedData')
        -- plan the actual work
        updates = chartStudySettings chartSettings <&> \cs ->
            -- check if thing is in hashmap
            let vec = case ccData HM.!? cs of
                    Nothing -> VS.replicate cTicks 0
                    Just a  -> a
                -- shift if neccessary
                vec' = if shiftNeccessary && cs `HM.member` ccData && shiftInterval < VS.length vec then
                          -- TODO: unsafeSlice && unsafeUpdate_ - see #16
                          let sliceLength = VS.length vec - shiftInterval
                          in VS.update_
                                (VS.update_ vec (VS.enumFromN 0 sliceLength) (VS.slice shiftInterval sliceLength vec))
                                (VS.enumFromN sliceLength (cTicks - sliceLength))
                                (VS.replicate sliceLength 0)
                       else
                        vec
            -- NOW:
            --  newRange is (rangeFrom, rangeTo) in seconds with 0 being last midnight
            --  vec' has that range mapped to indices [0,cTicks-1] with every cRes being one time-slice
            --  everything in the chunkedChart-list has to be inserted if in range using the same time-notation like the range
            --  (hint: chunkedChart is (time, [data]), with time being the point at the end of each contained interval)
                findOLHC   as = fromJust $ find (\case OLHC{}   -> True; _ -> False) as
                findVolume as = fromJust $ find (\case Volume{} -> True; _ -> False) as
            in case cs of
              ChartStudyTypeDirect -> HM.insert cs (vec' VS.// ((\(i, ChartPoint _ x _ _) -> (i,x))                       <$> foldedData))
              ChartStudyTypeOpen   -> HM.insert cs (vec' VS.// ((\(i, ChartPoint _ _ _ s) -> (i,olhc_open  $ findOLHC s)) <$> foldedData))
              ChartStudyTypeLow    -> HM.insert cs (vec' VS.// ((\(i, ChartPoint _ _ _ s) -> (i,olhc_low   $ findOLHC s)) <$> foldedData))
              ChartStudyTypeHigh   -> HM.insert cs (vec' VS.// ((\(i, ChartPoint _ _ _ s) -> (i,olhc_high  $ findOLHC s)) <$> foldedData))
              ChartStudyTypeClose  -> HM.insert cs (vec' VS.// ((\(i, ChartPoint _ _ _ s) -> (i,olhc_close $ findOLHC s)) <$> foldedData))
              ChartStudyTypeSMA x  -> id --TODO: implement #13
              ChartStudyTypeVolume -> HM.insert cs (vec' VS.// ((\(i, ChartPoint _ _ _ s) -> (i,volume $ findVolume s))   <$> foldedData))
    let ccData' = foldl' (.) id updates ccData
    logDebug $ displayShow ccData'
    let fillRange = case traceShowCommentId' "(ccFill, lastDataPoint)" (ccFill,lastDataPoint) of
                                          -- no new data, but different range. shift along.
                (Just (mi,ma), Nothing) -> if ma - shiftInterval < 0 then Nothing else Just (max 0 $ mi - shiftInterval, ma - shiftInterval)
                                          -- no data saved yet
                (Nothing, Nothing)      -> Nothing
                                          -- new data, update fillRange
                (mima, Just lastDP)     -> let lastBlock = let d = timeOfDay lastDP in (intervalTo d + intervalFrom d) `div` 2
                               in case timePointToIndex' $ lastBlock of
                                    Left err -> error $ "impossible #1 Chart.hs - " <> err
                                    Right ma -> if ma < cTicks
                                      then case mima of  -- check if we rotated out or have no fill-range set.
                                                Nothing        ->                                 Just (max 0 $ ma - shiftInterval, ma)
                                                Just (mi, ma') -> if ma' - shiftInterval < 0 then Just (max 0 $ ma - shiftInterval, ma)
                                                                                             else Just (max 0 $ mi - shiftInterval, ma)
                                      else error $ "impossible #2 Chart.hs - " <> show ma <> ">=" <> show cTicks
    logDebug $ displayShow fillRange
    return (Just $ TimePoint lUpdate, ChartCacheData ccData' ccAxis' (Just newRange) fillRange)

getChunkedDay :: Chart -> Maybe Int -> (Maybe TimePoint, [ChartPoint])
getChunkedDay Chart{..} chunkResolution = case toChartData [] chunkedData of
                                            (Nothing, x) -> (Nothing, x)
                                            (Just ChartPoint{..}, x)
                                              | timeOfDay == mempty -> (Nothing, x)
                                              | otherwise -> (Just $ TimePoint $ timeIntervalEnd timeOfDay, x)
    where
        chunkedData = filter (not . null . snd) $ chunkChart 0 86400 (fromMaybe (chartCacheResolution chartCacheSettings) chunkResolution) chartData

indexToTimePoint :: Maybe ChartCacheSettings -> TimeInterval -> Int -> TimePoint
indexToTimePoint ccs (TimeInterval rFrom rTo) i = TimePoint $ rFrom + i*cRes
  where
    ChartCacheSettings cRes _ = fromMaybe defChartCacheSettings ccs

timePointToIndex :: Maybe ChartCacheSettings -> TimeInterval -> TimePoint -> Either String Int
timePointToIndex ccs (TimeInterval rFrom rTo) (TimePoint p) = if
    | p < rFrom || p > rTo           -> Left $ "timePointToIndex " <> show ((cRes, cTicks),(rFrom, rTo),TimePoint p) <> ": " <> show p <> " out of range."
    | result < 0 || result >= cTicks -> Left $ "timePointToIndex " <> show ((cRes, cTicks),(rFrom, rTo),TimePoint p) <> ": " <> show result <> " is invalid index"
    | otherwise                      -> Right result
  where
    ChartCacheSettings cRes cTicks = fromMaybe defChartCacheSettings ccs
    result = (p - rFrom) `div` cRes
